@@ -10,6 +10,8 @@ module Address = struct
 
   external create : string -> string -> t = "caml_lo_address_new"
   let create host port = create host (string_of_int port)
+
+  external default : unit -> t = "caml_lo_address_default"
 end
 
 module Message = struct
@@ -19,7 +21,7 @@ module Message = struct
 
   type data =
     [
-      `Int32 of int
+    | `Int32 of int
     | `Float of float
     | `String of string
     | `Blob of string
@@ -43,6 +45,23 @@ module Message = struct
     List.iter (add m) d
 
   external send : Address.t -> string -> t -> unit = "ocaml_lo_send_message"
+
+  let to_string (x:data) =
+    match x with
+    | `Int32 n -> string_of_int n
+    | `Int64 n -> string_of_int n
+    | `Float f -> string_of_float f
+    | `Double f -> string_of_float f
+    | `Blob b -> b
+    | `Timetag (t1,t2) -> Printf.sprintf "(%d,%d)" t1 t2
+    | `String s -> s
+    | `Symbol s -> s
+    | `Char c -> String.make 1 c
+    | `Midi _ -> "<MIDI>"
+    | `True -> "true"
+    | `False -> "false"
+    | `Nil -> "nil"
+    | `Infinitum -> "inf"
 end
 
 let send addr path data =
@@ -53,8 +72,15 @@ let send addr path data =
 module Server = struct
   type t
 
+  exception Stopped
+
+  let () =
+    Callback.register_exception "lo_exn_stopped" Stopped;
+
   external create : string -> (string -> Message.data array -> unit) -> t = "caml_lo_server_new"
   let create p h = create (string_of_int p) h
 
   external recv : t -> unit = "caml_lo_server_recv"
+
+  external stop : t -> unit = "caml_lo_server_free"
 end
